import sys
import argparse

import numpy as np
import skimage.external.tifffile as tiff

from .normxcorr import normxcorr2_fftw
from .inputfile import InputFile


def to_dtype(x, dtype):
    x = np.rint(x) if np.issubdtype(dtype, np.integer) else x
    return x.astype(dtype, copy=False)


def stitch(aname, bname, z_frame, axis, overlap, max_shift_z=20,
           max_shift_y=150, max_shift_x=20):
    """Compute optimal shift between adjacent tiles.

    Two 3D tiles are compared at the specified frame index to find their best
    alignment. The following naming conventions are used:

    * Z is the direction along the stack height,
    * (X, Y) is the frame plane,
    * Y is the direction along which frames are supposed to overlap,
    * X is the direction orthogonal to Y in the frame plane (X, Y).

    Parameters
    ----------
    aname : str
        Input file name.
    bname : str
        Input file name.
    z_frame : int
        Index of frame used for alignment.
    axis : int
        Perform stitching along this axis. Following the convention above, this
        axis will be called Y. The following values can be used:

        1. stitch vertically
        2. stitch horizontally
    overlap : int
        Overlap height in px along the stitching axis (Y).
    max_shift_z : int
        Maximum allowed shift in px along the stack height (Z).
    max_shift_y : int
        Maximum allowed shift in px *of the overlapping area* along (Y).
    max_shift_x : int
        Maximum allowed lateral shift in px.

    Returns
    -------
    tuple
        Optimal shifts and overlap score computed by means of normalized
        cross correlation: [`dz`, `dy`, `dx`, `score`].
    """
    a = InputFile(aname)
    b = InputFile(bname)

    a.channel = -2
    b.channel = -2

    z_min = z_frame - max_shift_z
    z_max = z_frame + max_shift_z + 1

    aslice = a.zslice(z_min, z_max)
    if axis == 2:
        aslice = np.rot90(aslice, axes=(-1, -2))
    a_roi = aslice[..., -(overlap + max_shift_y):, :]

    bframe = b.zslice_idx(z_frame)
    if axis == 2:
        bframe = np.rot90(bframe, axes=(-1, -2))
    b_roi = bframe[..., :overlap - max_shift_y, max_shift_x:-max_shift_x]

    tiff.imsave('aslice.tiff', a_roi.astype(np.float32))
    tiff.imsave('bframe.tiff', b_roi.astype(np.float32))

    xcorr = normxcorr2_fftw(a_roi, b_roi)
    tiff.imsave('xcorr.tiff', xcorr.astype(np.float32))

    shift = list(np.unravel_index(np.argmax(xcorr), xcorr.shape))
    score = xcorr[tuple(shift)]

    print('shift: ' + str(shift))
    z_a = shift[0]
    shift[0] -= max_shift_z
    shift[1] = overlap + max_shift_y - shift[1]
    shift[2] -= max_shift_x

    print('max @ {}: {}, score: {:.3}'.format(z_frame, shift, score))

    a_roi = aslice[z_a, -shift[1]:, ...]
    b_roi = np.squeeze(bframe)[0:shift[1], ...]

    dx = shift[2]
    if dx > 0:
        a_roi = np.pad(a_roi, ((0, 0), (0, dx)), mode='constant')
        b_roi = np.pad(b_roi, ((0, 0), (dx, 0)), mode='constant')
    else:
        dx *= -1
        a_roi = np.pad(a_roi, ((0, 0), (dx, 0)), mode='constant')
        b_roi = np.pad(b_roi, ((0, 0), (0, dx)), mode='constant')

    rad = np.linspace(0.0, np.pi, a_roi.shape[0], dtype=np.float32)
    alpha = (np.cos(rad) + 1) / 2
    alpha = alpha[:, np.newaxis]

    fused = to_dtype(a_roi * alpha + b_roi * (1 - alpha), aslice.dtype)
    if axis == 2:
        fused = np.rot90(fused, axes=(-1, -2), k=3)

    tiff.imsave('fused.tiff', fused.astype(np.float32))
    return shift + [score]


def main():
    parser = argparse.ArgumentParser(
        description='Stitch a pair of files',
        epilog='Author: Giacomo Mazzamuto <mazzamuto@lens.unifi.it>',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('input_file1', help='input file (.dcimg)')
    parser.add_argument('input_file2', help='input file (.dcimg)')

    parser.add_argument('-f', type=int, dest='frame', required=True,
                        help='frame of input_file2')
    parser.add_argument('-a', type=int, help='axis (1=Y, 2=X)', dest='axis',
                        choices=(1, 2), default=1)
    parser.add_argument('--overlap', type=int, default=100, help='overlap')
    parser.add_argument('--Mz', type=int, default=20, help='Max shift along Z')
    parser.add_argument('--My', type=int, default=100,
                        help='Max shift along Y')
    parser.add_argument('--Mx', type=int, default=20,
                        help='Max shift along X (lateral)')

    args = parser.parse_args(sys.argv[1:])

    stitch(args.input_file1, args.input_file2, z_frame=args.frame,
           overlap=args.overlap, axis=args.axis, max_shift_z=args.Mz,
           max_shift_y=args.My, max_shift_x=args.Mx)


if __name__ == '__main__':
    main()



import glob
import os.path

import numpy as np
import skimage.external.tifffile as tiff


class TiffWrapper(object):
    def __init__(self, file_name=None):
        self.file_name = file_name

        self.tfile = None
        self.flist = None
        self.glob_mode = False

        if file_name is not None:
            self.open()

    @property
    def nfrms(self):
        nfrms = len(self.tfile.pages)
        if self.glob_mode:
            nfrms *= len(self.flist)
        return nfrms

    @property
    def xsize(self):
        return self.tfile.pages[0].image_width

    @property
    def ysize(self):
        return self.tfile.pages[0].image_length

    @property
    def axes(self):
        return self.tfile.pages[0].axes

    @property
    def nchannels(self):
        if self.axes.startswith('YX'):
            return self.tfile.pages[0]._shape[-1]
        elif self.axes == 'SYX':
            return self.tfile.pages[0]._shape[1]

    @property
    def dtype(self):
        return np.dtype(self.tfile.pages[0].dtype)

    @property
    def shape(self):
        if self.nchannels > 1:
            return (self.nfrms, self.nchannels, self.ysize, self.xsize)
        else:
            return (self.nfrms, self.ysize, self.xsize)

    def open(self, file_name=None):
        if file_name is not None:
            self.file_name = file_name

        if os.path.isdir(self.file_name):
            self.glob_mode = True
            flist = []
            flist += glob.glob(os.path.join(self.file_name, '*.tif*'))
            flist += glob.glob(os.path.join(self.file_name, '*.TIF*'))
            flist = sorted(flist)
            fname = flist[0]
            self.flist = flist
        else:
            self.glob_mode = False
            fname = self.file_name

        self.tfile = tiff.TiffFile(fname)
        setattr(self, 'close', getattr(self.tfile, 'close'))

    def zslice(self, start_frame, end_frame=None, dtype=None, copy=True):
        if end_frame is None:
            end_frame = start_frame + 1

        if not self.glob_mode:
            a = self.tfile.asarray(slice(start_frame, end_frame))
        else:
            frames_per_file = self.nfrms // len(self.flist)
            start_file = start_frame // frames_per_file
            end_file = end_frame // frames_per_file
            a = tiff.imread(self.flist[start_file:end_file], pattern='')

        if end_frame - start_frame == 1:
            a = np.expand_dims(a, axis=0)

        if self.axes == 'SYX':
            a = np.moveaxis(a, 1, -1)

        if dtype is None:
            return a
        return a.astype(dtype)



import queue
import logging
import argparse
import threading

import json
import yaml

import coloredlogs

import numpy as np
import pandas as pd

from .inputfile import InputFile
from .filematrix import FileMatrix
from .normxcorr import normxcorr2_fftw

from .version import __version__


logger = logging.getLogger(__name__)
coloredlogs.install(level='DEBUG', fmt='%(levelname)s [%(name)s]: %(message)s')


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


def parse_args():
    parser = argparse.ArgumentParser(
        description='''
Stitch tiles in a folder.

The following naming conventions are used:
* Z is the direction along the stack height,
* (X, Y) is the frame plane,
* y is the direction along which frames are supposed to overlap,
* x is the direction orthogonal to y in the frame plane (X, Y).

Unless otherwise stated, all values are expected in px.
    ''',
        epilog='Author: Giacomo Mazzamuto <mazzamuto@lens.unifi.it>\n'
               'Version: {}'.format(__version__),
        formatter_class=CustomFormatter)

    parser.add_argument('input_folder', help='input folder')
    parser.add_argument('-o', type=str, default='stitch.yml',
                        dest='output_file', help='output file')
    parser.add_argument('-c', type=str, default='s', dest='channel',
                        choices=['r', 'g', 'b', 's'], help='color channel')
    parser.add_argument('-n', type=int, default=8, dest='n_of_threads',
                        help='number of parallel threads to use')

    group = parser.add_argument_group(
        'pixel size', 'If specified, the corresponding options can be '
                      'expressed in your custom units.')
    group.add_argument('--px-size-xy', type=float, default=1,
                       help='pixel size in the (X, Y) plane')
    group.add_argument('--px-size-z', type=float, default=1,
                       help='pixel size in the Z direction')

    group = parser.add_argument_group('maximum shifts')
    group.add_argument('--dz', type=float, required=True, dest='max_dz',
                       help='maximum allowed shift along Z')

    group.add_argument('--dy', type=float, required=True, dest='max_dy',
                       help='maximum allowed shift along y (the stitching '
                            'axis) relative to the nominal overlap')

    group.add_argument('--dx', type=float, required=True, dest='max_dx',
                       help='maximum allowed shift along x (lateral shift)')

    group = parser.add_argument_group('overlaps')
    group.add_argument('--overlap-h', type=float, required=True, metavar='OH',
                       help='nominal overlap along the horizontal axis')

    group.add_argument('--overlap-v', type=float, required=True, metavar='OV',
                       help='nominal overlap along the vertical axis')

    group = parser.add_argument_group(
        'multiple sampling along Z',
        description='Measure the optimal shift at different heights around '
                    'the center of the stack')

    group.add_argument('--z-samples', type=int, default=1, metavar='ZSAMP',
                       help='number of samples to take along Z')

    group.add_argument('--z-stride', type=float, default=None,
                       help='stride used for multiple Z sampling')

    group = parser.add_argument_group('tile ordering')
    group.add_argument('--iX', action='store_true', dest='invert_x',
                       help='invert tile ordering along X')
    group.add_argument('--iY', action='store_true', dest='invert_y',
                       help='invert tile ordering along Y')

    args = parser.parse_args()

    channels = {
        's': -2,  # sum
        'r': 0,
        'g': 1,
        'b': 2
    }

    args.channel = channels[args.channel]

    if args.z_samples > 1 and args.z_stride is None:
        args.z_stride = args.dz * 1.2

    args.max_dx = int(round(args.max_dx / args.px_size_xy))
    args.max_dy = int(round(args.max_dy / args.px_size_xy))
    args.max_dz = int(round(args.max_dz / args.px_size_z))
    if args.z_stride is not None:
        args.z_stride = int(round(args.z_stride / args.px_size_z))
    else:
        args.z_stride = 0

    args.overlap_v = int(round(args.overlap_v / args.px_size_xy))
    args.overlap_h = int(round(args.overlap_h / args.px_size_xy))

    setattr(args, 'ascending_tiles_x', not args.invert_x)
    setattr(args, 'ascending_tiles_y', not args.invert_y)

    return args


class Runner(object):
    def __init__(self):
        self.channel = None
        self.q = None
        self.output_q = None
        self.data_queue = None
        self.initial_queue_length = None
        self.input_folder = None
        self.output_file = None
        self.z_samples = None
        self.z_stride = None
        self.overlap_v = None
        self.overlap_h = None
        self.max_dx = None
        self.max_dy = None
        self.max_dz = None
        self.compute_average = False
        self.ascending_tiles_x = True
        self.ascending_tiles_y = True
        self.df = None
        self.fm = None
        self.px_size_xy = 1
        self.px_size_z = 1
        self.n_of_threads = 1

    @property
    def overlap_dict(self):
        return {1: self.overlap_v, 2: self.overlap_h}

    def initialize_queue(self):
        fm = FileMatrix(
            self.input_folder, self.ascending_tiles_x, self.ascending_tiles_y)
        self.fm = fm

        stitch_X = {
            'axis': 2,
            'sort': ['Z', 'X', 'Y'],
            'groupby': 'Y'
        }

        stitch_Y = {
            'axis': 1,
            'sort': ['Z', 'Y', 'X'],
            'groupby': 'X'
        }

        q = queue.Queue()

        for s in fm.slices:
            df = fm.data_frame.loc[list(s.nodes())]
            for stitch_config in [stitch_X, stitch_Y]:
                view = df.sort_values(stitch_config['sort'],
                                      ascending=True)
                view = view.groupby(stitch_config['groupby'])

                for name, group in view:
                    Z_from = max(group['Z'])
                    Z_to = min(group['Z_end'])
                    tiles = list(group.itertuples())
                    atile = tiles[0]
                    for btile in tiles[1::]:
                        central_frame = (Z_to - Z_from) // 2
                        start_frame = (
                            central_frame
                            - (self.z_samples // 2 * self.z_stride)
                            + (0 if self.z_samples % 2 else self.z_stride // 2)
                        )
                        z_frames = []
                        for i in range(0, self.z_samples):
                            z_frames.append(start_frame + i * self.z_stride)
                        params_dict = {
                            'aname': atile.Index,
                            'bname': btile.Index,
                            'z_frames': z_frames,
                            'axis': stitch_config['axis'],
                        }
                        q.put(params_dict)
                        atile = btile
        self.q = q

    def worker(self, initial_queue_length):
        while True:
            item = self.data_queue.get()
            if item is None:
                break
            try:
                aname = item[0]
                bname = item[1]
                axis = item[2]
                aslice = item[3]
                bframe = item[4]
                z_frame = item[5]

                xcorr = normxcorr2_fftw(aslice, bframe)

                shift = list(np.unravel_index(np.argmax(xcorr), xcorr.shape))
                score = xcorr[tuple(shift)]
                if score < 0 or score > 1:
                    score = 0

                progress = 100 * (1 - self.q.qsize() / initial_queue_length)
                logger.info('{progress:.2f}%\t{aname}\t{bname}\t{z_frame}\t'
                            '{shift}\t{score:.3f}'.format(
                    progress=progress, aname=aname, bname=bname,
                    z_frame=z_frame, shift=shift, score=score))
                self.output_q.put(
                    [aname, bname, axis, z_frame] + shift + [score])
            finally:
                self.data_queue.task_done()

    def keep_filling_data_queue(self):
        while True:
            try:
                item = self.q.get_nowait()
            except queue.Empty:
                break
            aname = item['aname']
            bname = item['bname']
            z_frames = item['z_frames']
            axis = item['axis']
            overlap = self.overlap_dict[axis]

            a = InputFile(aname)
            b = InputFile(bname)

            a.channel = self.channel
            b.channel = self.channel

            for z_frame in z_frames:
                z_min = z_frame - self.max_dz
                z_max = z_frame + self.max_dz + 1

                aslice = a.zslice(z_min, z_max, copy=True)
                if axis == 2:
                    aslice = np.rot90(aslice, axes=(-1, -2))
                aslice = aslice[..., -(overlap + self.max_dy):, :]

                bframe = b.zslice_idx(z_frame, copy=True)
                if axis == 2:
                    bframe = np.rot90(bframe, axes=(-1, -2))
                bframe = bframe[..., :overlap - self.max_dy,
                                self.max_dx:-self.max_dx]

                aslice = aslice.astype(np.float32)
                bframe = bframe.astype(np.float32)

                self.data_queue.put(
                    [aname, bname, axis, aslice, bframe, z_frame])

            self.q.task_done()

    def run(self):
        self.initialize_queue()
        self.data_queue = queue.Queue(maxsize=int(self.n_of_threads * 2))
        self.output_q = queue.Queue()
        threads = []
        for i in range(self.n_of_threads):
            t = threading.Thread(target=self.worker, args=(self.q.qsize(),))
            t.start()
            threads.append(t)

        self.keep_filling_data_queue()

        # block until all tasks are done
        self.data_queue.join()

        # stop workers
        for i in range(self.n_of_threads):
            self.data_queue.put(None)
        for t in threads:
            t.join()

        df = pd.DataFrame(list(self.output_q.queue))
        df.columns = ['aname', 'bname', 'axis', 'z_frame', 'dz', 'dy', 'dx',
                      'score']
        self.df = df

        self.save_results_to_file()

        df[['dx_px', 'dy_px', 'dz_px']] = df[['dx', 'dy', 'dz']]
        df[['dx', 'dy']] *= self.px_size_xy
        df['dz'] *= self.px_size_z

        cols = ['dx', 'dy', 'dz', 'score', 'dx_px', 'dy_px', 'dz_px']
        print(df[cols].describe())

    def save_results_to_file(self):
        self.fm.save_to_yaml(self.output_file, 'w')

        attrs = ['max_dx', 'max_dy', 'max_dz', 'overlap_v', 'overlap_h',
                 'ascending_tiles_x', 'ascending_tiles_y', 'px_size_xy',
                 'px_size_z', 'z_samples', 'z_stride']

        options = {}
        for attr in attrs:
            options[attr] = getattr(self, attr)

        with open(self.output_file, 'a') as f:
            yaml.dump(
                {
                    'xcorr-options': options,
                    'xcorr': json.loads(self.df.to_json(orient='records'))
                }, f, default_flow_style=False)


def main():
    arg = parse_args()

    r = Runner()

    keys = ['input_folder', 'output_file', 'channel', 'max_dx', 'max_dy',
            'max_dz', 'z_samples', 'z_stride', 'overlap_v', 'overlap_h',
            'ascending_tiles_x', 'ascending_tiles_y', 'px_size_xy',
            'px_size_z', 'n_of_threads']

    for key in keys:
        setattr(r, key, getattr(arg, key))

    r.run()


if __name__ == '__main__':
    main()



import sys
import os.path
import logging
import argparse

import yaml
import coloredlogs

from ..version import __version__

from . import absolute_positions
from .fuse_runner import FuseRunner
from ..filematrix import FileMatrix
from .xcorr_filematrix import XcorrFileMatrix
from .global_optimization import absolute_position_global_optimization

logger = logging.getLogger(__name__)
coloredlogs.install(level='DEBUG', fmt='%(levelname)s [%(name)s]: %(message)s')


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


def parse_args():
    parser = argparse.ArgumentParser(
        description='Fuse stitched tiles in a folder.',
        epilog='Author: Giacomo Mazzamuto <mazzamuto@lens.unifi.it>\n'
               'Version: {}'.format(__version__),
        formatter_class=CustomFormatter)

    parser.add_argument(
        'yml_file',
        help='.yml file produced by stitch align. It will also be used for '
             'saving absolute coordinates. If a directory is specified'
             'instead of a file, uses a file named "stitch.yml" if present. '
             'If the file does not exist, it will be created (only where '
             'applicable: see option -s).')

    group = parser.add_argument_group('output')
    group.add_argument('-o', type=str, dest='output_filename',
                       help='output file name. If not specified, no tiff '
                            'output is produced, only absoulute coordinates '
                            'are computed.')

    group.add_argument('-d', dest='debug', action='store_true',
                       help='overlay debug info')

    group.add_argument('-c', type=int, default=-1, dest='channel',
                       help='channel')

    group.add_argument('--zmin', type=float, default=0)
    group.add_argument('--zmax', type=float, default=None, help='noninclusive')

    group = parser.add_argument_group(
        'absolute positions', 'by default, absolute positions are computed by '
                              'taking the maximum score in cross correlations')
    me_group = group.add_mutually_exclusive_group()
    me_group.add_argument('-m', dest='abs_mode', action='store_const',
                          const='maximum_score',
                          help='take the maximum score in cross correlations '
                               '(default)')

    me_group.add_argument('-a', dest='abs_mode', action='store_const',
                          const='weighted_average',
                          help='take the average result weighted by the score')

    me_group.add_argument('-s', dest='abs_mode', action='store_const',
                          const='nominal_positions',
                          help='use nominal stage positions')

    group.add_argument('-f', action='store_true', default=False,
                       dest='force_recomputation',
                       help='force recomputation of absolute positions')

    group.add_argument('--no-global', action='store_true',
                       help='do not perform global optimization (where '
                            'applicable)')

    group = parser.add_argument_group('tile ordering (option -s only)')
    group.add_argument('--iX', action='store_true', dest='invert_x',
                       help='invert tile ordering along X')
    group.add_argument('--iY', action='store_true', dest='invert_y',
                       help='invert tile ordering along Y')

    group = parser.add_argument_group(
        'pixel size', 'If specified, the corresponding options can be '
                      'expressed in your custom units.')
    group.add_argument('--px-size-xy', type=float,
                       help='pixel size in the (X, Y) plane')
    group.add_argument('--px-size-z', type=float,
                       help='pixel size in the Z direction')

    args = parser.parse_args()

    return args


def main():
    args = parse_args()

    if os.path.isdir(args.yml_file):
        temp = os.path.join(args.yml_file, 'stitch.yml')
        if os.path.exists(temp):
            args.yml_file = temp

    asc_keys = ['ascending_tiles_x', 'ascending_tiles_y']
    for k in asc_keys:
        setattr(args, k, None)

    # replace None args with values found in yml file
    old_abs_mode = None
    if os.path.isfile(args.yml_file):
        with open(args.yml_file, 'r') as f:
            y = yaml.load(f)
            try:
                old_abs_mode = y['fuser-options']['abs_mode']
            except KeyError:
                pass
        keys = ['px_size_z', 'px_size_xy'] + asc_keys
        for k in keys:
            if getattr(args, k) is None:
                try:
                    setattr(args, k, y['xcorr-options'][k])
                except KeyError:
                    pass

    using_old_abs_mode = False
    if args.abs_mode is None:
        args.abs_mode = old_abs_mode
        using_old_abs_mode = True
    if args.abs_mode is None:
        args.abs_mode = 'maximum_score'

    if not os.path.isfile(args.yml_file):
        if args.abs_mode != 'nominal_positions':
            logger.error("No stitch file specified or found. Please specify "
                         "input file or run with -s.")
            sys.exit(1)

    for k in ['x', 'y']:
        temp_k = 'ascending_tiles_' + k
        if getattr(args, temp_k, None) is None:
            setattr(args, temp_k, not getattr(args, 'invert_' + k))

    attrs = ['px_size_z', 'px_size_xy']
    for a in attrs:
        if getattr(args, a, None) is None:
            if args.abs_mode == 'nominal_positions' and not using_old_abs_mode:
                logger.error(
                    "px sizes need to be specified when using option -s")
                sys.exit(1)
            else:
                setattr(args, a, 1)

    args.zmin = int(round(args.zmin / args.px_size_z))
    if args.zmax is not None:
        args.zmax = int(round(args.zmax / args.px_size_z))

    # =========================================================================
    # init FileMatrix
    # =========================================================================
    fm = FileMatrix(args.yml_file,
                    ascending_tiles_x=args.ascending_tiles_x,
                    ascending_tiles_y=args.ascending_tiles_y)

    if args.force_recomputation or args.abs_mode != old_abs_mode:
        fm.clear_absolute_positions()

    cols = fm.data_frame.columns
    logger.info('absolute positions mode: {}'.format(args.abs_mode))
    if 'Xs' in cols and 'Ys' in cols and 'Zs' in cols:
        logger.info('using absolute positions from {}'.format(args.yml_file))
    elif args.abs_mode == 'nominal_positions':
        fm.compute_nominal_positions(args.px_size_z, args.px_size_xy)
    else:
        xcorr_fm = XcorrFileMatrix()
        xcorr_fm.load_yaml(fm.input_path)
        compute_average = \
            True if args.abs_mode == 'weighted_average' else False
        xcorr_fm.aggregate_results(compute_average=compute_average)

        sdf = xcorr_fm.stitch_data_frame

        absolute_positions.compute_shift_vectors(fm.data_frame, sdf)
        absolute_positions.compute_initial_guess(fm.data_frame, sdf)

        if not args.no_global:
            absolute_position_global_optimization(fm.data_frame, sdf,
                                                  xcorr_fm.xcorr_options)

    # =========================================================================
    # init FuseRunner
    # =========================================================================
    if args.output_filename is not None:
        fr = FuseRunner(fm)

        keys = ['zmin', 'zmax', 'output_filename', 'debug', 'channel']

        for k in keys:
            setattr(fr, k, getattr(args, k))

        logger.info("invert X: {}, invert Y: {}".format(
            not args.ascending_tiles_x, not args.ascending_tiles_y))
        logger.info("output shape: {}".format(fr.output_shape))
        logger.info("voxel size (ZYX): {} * {} * {}".format(
            args.px_size_z, args.px_size_xy, args.px_size_xy))

        fr.run()
    else:
        logger.warning("No output file specified.")

    if os.path.isdir(args.yml_file):
        args.yml_file = os.path.join(args.yml_file, 'stitch.yml')
        fm.save_to_yaml(args.yml_file, 'w')
    else:
        fm.save_to_yaml(args.yml_file, 'update')

    with open(args.yml_file, 'r') as f:
        y = yaml.load(f)
    fr_options = {}
    keys = ['abs_mode', 'px_size_xy', 'px_size_z', 'ascending_tiles_x',
            'ascending_tiles_y']
    for k in keys:
        fr_options[k] = getattr(args, k)
    y['fuser-options'] = fr_options

    with open(args.yml_file, 'w') as f:
        yaml.dump(y, f, default_flow_style=False)


if __name__ == '__main__':
    main()



import os.path

import numpy as np

try:
    import dcimg
except ImportError:
    pass

from .ffmpeg_wrapper import FFMPEGWrapper
from .tiffwrapper import TiffWrapper


class InputFile(object):
    def __init__(self, file_name=None):
        self.file_name = file_name
        self.wrapper = None
        self._channel = -1
        self.nchannels = 1

        self.nfrms = None

        if file_name is not None:
            self.open()

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def __getitem__(self, item):
        item = np.index_exp[item]  # ensure item is a tuple

        # ensure all items are slice objects
        myitem = []
        for i in item:
            if isinstance(i, int):
                start = i
                stop = i + 1
                step = 1
            elif i is Ellipsis:
                for _ in range(0, len(self.shape) - len(item) + 1):
                    myitem.append(slice(0, self.shape[len(myitem)], 1))
                continue
            elif isinstance(i, slice):
                start = i.start
                stop = i.stop
                step = i.step if i.step is not None else 1
            else:
                raise TypeError("Invalid type: {}".format(type(i)))

            curr_max = self.shape[len(myitem)]
            if start is None:
                start = 0 if step > 0 else curr_max
            elif start < 0:
                start += curr_max
            elif start > curr_max:
                start = curr_max

            if stop is None:
                stop = curr_max if step > 0 else 0
            elif stop < 0:
                stop += curr_max
            elif stop > curr_max:
                stop = curr_max

            if step < 0:
                start += abs(step)
                stop += abs(step)

            myitem.append(slice(start, stop, step))

        for _ in range(0, len(self.shape) - len(myitem)):
            myitem.append(slice(0, self.shape[len(myitem)], 1))

        try:
            return self.wrapper.__getitem__(myitem)
        except AttributeError:
            pass

        a = self.zslice(myitem[0].start, myitem[0].stop)

        myitem[0] = slice(0, None, myitem[0].step)
        a = a[myitem]

        return np.squeeze(a)

    @property
    def channel(self):
        return self._channel

    @channel.setter
    def channel(self, value):
        if not self.nchannels > 1:
            return
        self._channel = value

    @property
    def file(self):
        return self.wrapper

    @file.setter
    def file(self, value):
        self.wrapper = value
        self._setattrs()

    @property
    def shape(self):
        """Shape of the whole image stack.

        Returns
        -------
        tuple
            (:attr:`nfrms`, `channels`, :attr:`xsize`, :attr:`ysize`) where
            `channels` is the number of color channels in the image. If
            :attr:`channel` is set or if there is only one channel,
            the `channels` dimension is squeezed.
        """
        s = [self.nfrms, self.xsize, self.ysize]
        if self.nchannels != 1 and self.channel == -1:
            s.insert(1, self.nchannels)
        return tuple(s)

    @property
    def file_size(self):
        return os.path.getsize(self.file_name)

    def open(self, file_name=None):
        if file_name is not None:
            self.file_name = file_name

        self._open()
        self._setattrs()

    def _open(self):
        try:
            self.wrapper = TiffWrapper(self.file_name)
            return
        except (FileNotFoundError, ValueError, IsADirectoryError):
            pass

        try:
            self.wrapper = dcimg.DCIMGFile(self.file_name)
            self.wrapper.retrieve_first_4_pixels = False
            return
        except (NameError, FileNotFoundError, ValueError, IsADirectoryError):
            pass

        try:
            self.wrapper = FFMPEGWrapper(self.file_name)
            return
        except ValueError:
            pass

        raise ValueError('Unsupported file type')

    def close(self):
        try:
            self.wrapper.close()
        except AttributeError:
            pass

    def _setattrs(self):
        l = ['nfrms', 'xsize', 'ysize', 'nchannels', 'dtype']

        for a in l:
            try:
                setattr(self, a, getattr(self.wrapper, a))
            except AttributeError:
                pass

    def zslice(self, start_frame, end_frame=None, dtype=None, copy=True):
        """Return a slice, i.e. a substack of frames.

        Parameters
        ----------
        start_frame : int
            first frame to select
        end_frame : int
            last frame to select (noninclusive). If None, defaults to
            :code:`start_frame + 1`
        dtype

        Returns
        -------
        :class:`numpy.ndarray`
            A numpy array of the original type or of `dtype`, if specified. The
            shape of the array is (`end_frame` - `start_frame`,
            :attr:`channels` :attr:`ysize`, :attr:`xsize`, :attr:`channels`)
            where `channels` is the number of color channels in the image. If
            :attr:`channel` is set or if there is only one channel, the
            `channels` dimension is squeezed.
        """
        l = self.wrapper.zslice(start_frame, end_frame, dtype, copy)
        if self.channel == -2:
            l = np.sum(l, axis=-1)
        elif self.channel != -1:
            l = l[..., self.channel]
        elif self.nchannels > 1:
            l = np.moveaxis(l, -1, -3)

        return l

    def zslice_idx(self, index, frames_per_slice=1, dtype=None, copy=True):
        """Return a slice, i.e. a substack of frames, by index.

        Parameters
        ----------
        index : int
            slice index
        frames_per_slice : int
            number of frames per slice
        dtype

        Returns
        -------
        :class:`numpy.ndarray`
            A numpy array, see :func:`slice`.
        """
        start_frame = index * frames_per_slice
        end_frame = start_frame + frames_per_slice
        return self.zslice(start_frame, end_frame, dtype, copy)

    def whole(self, dtype=None, copy=True):
        """Convenience function to retrieve the whole stack.

        Equivalent to call :func:`slice_idx` with `index` = 0 and
        `frames_per_slice` = :attr:`nfrms`

        Parameters
        ----------
        dtype

        Returns
        -------
        :class:`numpy.ndarray`
            A numpy array, see :func:`slice`.
        """
        return self.zslice_idx(0, self.nfrms, dtype, copy)

    def frame(self, index, dtype=None, copy=True):
        """Convenience function to retrieve a single slice.

        Same as calling :func:`slice` and squeezing.

        Parameters
        ----------
        index : int
            slice index
        dtype

        Returns
        -------
        :class:`numpy.ndarray`
            A numpy array, see :func:`slice`.
        """
        return np.squeeze(self.zslice_idx(index, dtype=dtype, copy=copy))



import pandas as pd


class Overlaps(object):
    def __init__(self, filematrix):
        self.fm = filematrix

        self.overlap_n = None
        self.overlap_s = None
        self.overlap_e = None
        self.overlap_w = None
        self.overlap_nw = None
        self.overlap_ne = None
        self.overlap_sw = None
        self.overlap_se = None

        self._compute_overlaps()

    def _compute_overlaps(self):
        def comp_diff(dest, row, other_name):
            other = fm_df.loc[other_name]
            temp = pd.Series()
            temp['Z_from'] = max(other['Zs'], row['Zs']) - row['Zs']
            temp['Z_to'] = min(other['Zs_end'], row['Zs_end']) - row['Zs']

            temp['Y_from'] = max(other['Ys'], row['Ys']) - row['Ys']
            temp['Y_to'] = min(other['Ys_end'], row['Ys_end']) - row['Ys']

            temp['X_from'] = max(other['Xs'], row['Xs']) - row['Xs']
            temp['X_to'] = min(other['Xs_end'], row['Xs_end']) - row['Xs']

            cols_to_zero(dest, row.name)

            if temp['Z_from'] > temp['Z_to'] or temp['Y_from'] > temp['Y_to'] \
                    or temp['X_from'] > temp['X_to']:
                pass
            else:
                dest.loc[row.name] = temp

        def cols_to_zero(dest, row_name):
            cols = ['Z_from', 'Z_to', 'Y_from', 'Y_to', 'X_from', 'X_to']
            for c in cols:
                dest.loc[row_name, c] = 0

        fm_df = self.fm.data_frame

        overlap_n = pd.DataFrame()
        overlap_s = pd.DataFrame()
        overlap_e = pd.DataFrame()
        overlap_w = pd.DataFrame()
        overlap_nw = pd.DataFrame()
        overlap_ne = pd.DataFrame()
        overlap_sw = pd.DataFrame()
        overlap_se = pd.DataFrame()

        for j in range(0, self.fm.Ny):
            for i in range(0, self.fm.Nx):
                row = self.fm.data_frame.loc[self.fm.name_array[j, i]]

                # north
                if j > 0:
                    parent = self.fm.name_array[j - 1, i]
                    comp_diff(overlap_n, row, parent)
                else:
                    cols_to_zero(overlap_n, row.name)

                # south
                if j < self.fm.Ny - 1:
                    parent = self.fm.name_array[j + 1, i]
                    comp_diff(overlap_s, row, parent)
                else:
                    cols_to_zero(overlap_s, row.name)

                # east
                if i < self.fm.Nx - 1:
                    parent = self.fm.name_array[j, i + 1]
                    comp_diff(overlap_e, row, parent)
                else:
                    cols_to_zero(overlap_e, row.name)

                # west
                if i > 0:
                    parent = self.fm.name_array[j, i - 1]
                    comp_diff(overlap_w, row, parent)
                else:
                    cols_to_zero(overlap_w, row.name)

                # north-west
                if i > 0 and j > 0:
                    parent = self.fm.name_array[j - 1, i - 1]
                    comp_diff(overlap_nw, row, parent)
                else:
                    cols_to_zero(overlap_nw, row.name)

                # north-east
                if i < self.fm.Nx - 1 and j > 0:
                    parent = self.fm.name_array[j - 1, i + 1]
                    comp_diff(overlap_ne, row, parent)
                else:
                    cols_to_zero(overlap_ne, row.name)

                # south-west
                if i > 0 and j < self.fm.Ny - 1:
                    parent = self.fm.name_array[j + 1, i - 1]
                    comp_diff(overlap_sw, row, parent)
                else:
                    cols_to_zero(overlap_sw, row.name)

                # south-east
                if i < self.fm.Nx - 1 and j < self.fm.Ny - 1:
                    parent = self.fm.name_array[j + 1, i + 1]
                    comp_diff(overlap_se, row, parent)
                else:
                    cols_to_zero(overlap_se, row.name)

        self.overlap_n = overlap_n.astype(int)
        self.overlap_s = overlap_s.astype(int)
        self.overlap_e = overlap_e.astype(int)
        self.overlap_w = overlap_w.astype(int)
        self.overlap_nw = overlap_nw.astype(int)
        self.overlap_ne = overlap_ne.astype(int)
        self.overlap_sw = overlap_sw.astype(int)
        self.overlap_se = overlap_se.astype(int)

    def __getitem__(self, tile_name):
        df = pd.DataFrame()
        names = ['n', 's', 'e', 'w', 'nw', 'ne', 'sw', 'se']
        for n in names:
            name = 'overlap_' + n
            overlap = getattr(self, name)
            df[n] = overlap.loc[tile_name]
        return df.transpose()



import logging
import os.path
import threading

from queue import Queue
from functools import lru_cache

import psutil
import numpy as np
import skimage.external.tifffile as tiff

from .fuse import fuse_queue, to_dtype
from .overlaps import Overlaps

from ..inputfile import InputFile

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class FuseRunner(object):
    def __init__(self, file_matrix):
        self.fm = file_matrix  #: :class:`FileMatrix`
        self.path, _ = os.path.split(self.fm.input_path)

        self.zmin = 0
        self.zmax = None
        self.debug = False
        self.output_filename = None
        self.channel = -1

        self._is_multichannel = None

    @property
    @lru_cache()
    def dtype(self):
        infile = os.path.join(self.path, self.fm.data_frame.iloc[0].name)
        with InputFile(infile) as f:
            return np.dtype(f.dtype)

    @property
    @lru_cache()
    def is_multichannel(self):
        if self.channel != -1:
            return False
        infile = os.path.join(self.path, self.fm.data_frame.iloc[0].name)
        with InputFile(infile) as f:
            if f.nchannels > 1:
                multichannel = True
            else:
                multichannel = False
        return multichannel

    @property
    def output_shape(self):
        thickness = self.fm.full_thickness
        if self.zmax is not None:
            thickness -= (thickness - self.zmax)
        thickness -= self.zmin

        infile = os.path.join(self.path, self.fm.data_frame.iloc[0].name)
        with InputFile(infile) as f:
            f.channel = self.channel
            output_shape = list(f.shape)

        output_shape[0] = thickness
        output_shape[-2] = self.fm.full_height
        output_shape[-1] = self.fm.full_width

        return tuple(output_shape)

    def run(self):
        ov = Overlaps(self.fm)

        total_byte_size = np.asscalar(np.prod(self.output_shape)
                                      * self.dtype.itemsize)
        bigtiff = total_byte_size > 2**31 - 1

        ram = psutil.virtual_memory().available

        # size in bytes of an xy plane (including channels) (float32)
        xy_size = np.asscalar(np.prod(self.output_shape[1::]) * 4)
        n_frames_in_ram = int(ram / xy_size / 1.5)

        n_loops = self.output_shape[0] // n_frames_in_ram

        partial_thickness = [n_frames_in_ram for _ in range(0, n_loops)]
        remainder = self.output_shape[0] % n_frames_in_ram
        if remainder:
            partial_thickness += [remainder]

        try:
            os.remove(self.output_filename)
        except FileNotFoundError:
            pass

        infile = os.path.join(self.path, self.fm.data_frame.iloc[0].name)
        with InputFile(infile) as f:
            f.channel = self.channel
            frame_shape = list(f.shape)[-2::]

        for thickness in partial_thickness:
            self.zmax = self.zmin + thickness
            fused = np.zeros(self.output_shape, dtype=np.float32)
            q = Queue(maxsize=20)

            t = threading.Thread(target=fuse_queue,
                                 args=(q, fused, frame_shape, self.debug))
            t.start()

            for index, row in self.fm.data_frame.iterrows():
                if self.zmax is None:
                    z_to = row.nfrms
                else:
                    z_to = self.zmax - row.Zs

                if z_to > row.nfrms:
                    z_to = row.nfrms

                if z_to <= 0:
                    continue

                z_from = self.zmin - row.Zs

                if z_from < 0:
                    z_from = 0

                if z_from >= z_to:
                    continue

                with InputFile(os.path.join(self.path, index)) as f:
                    f.channel = self.channel
                    logger.info(
                        'opening {}\tz=[{}:{}]'.format(index, z_from, z_to))
                    zslice = f.zslice(
                        z_from, z_to, dtype=np.float32, copy=True)

                top_left = [row.Zs + z_from - self.zmin, row.Ys, row.Xs]
                overlaps = ov[index].copy()
                overlaps = overlaps.loc[
                    (overlaps['Z_from'] <= z_to) & (overlaps['Z_to'] >= z_from)
                ]

                overlaps['Z_from'] -= z_from
                overlaps['Z_to'] -= z_from

                overlaps.loc[overlaps['Z_from'] < 0, 'Z_from'] = 0

                q.put([zslice, index, z_from, None, top_left, overlaps])

            q.put([None, None, None, None, None, None])  # close queue

            t.join()  # wait for fuse thread to finish
            print('=================================')

            if self.is_multichannel:
                fused = np.moveaxis(fused, -3, -1)

            fused = to_dtype(fused, self.dtype)
            logger.info('saving output to {}'.format(self.output_filename))
            tiff.imsave(self.output_filename, fused, append=True,
                        bigtiff=bigtiff)

            self.zmin += thickness



import json
import subprocess as sp

import numpy as np


class FFMPEGWrapper(object):
    def __init__(self, file_name=None):
        self.file_name = file_name

        self.proc = None

        self._probed_dict = None

        if file_name is not None:
            self.open()

    @property
    def nfrms(self):
        return int(self._probed_dict['streams'][0]['nb_frames'])

    @property
    def xsize(self):
        return self._probed_dict['streams'][0]['width']

    @property
    def ysize(self):
        return self._probed_dict['streams'][0]['height']

    @property
    def pix_fmt(self):
        return self._probed_dict['streams'][0]['pix_fmt']

    @property
    def nchannels(self):
        pix_fmt = self.pix_fmt
        if 'gray' in pix_fmt or 'mono' in pix_fmt or 'pal8' in pix_fmt:
            return 1
        elif 'yuva' in pix_fmt:
            return 4
        elif 'yuv' in pix_fmt or 'rgb' in pix_fmt or 'gbr' in pix_fmt or \
                'bgr' in pix_fmt:
            return 3

    @property
    def dtype(self):
        return np.uint8

    @property
    def shape(self):
        s = (self.nfrms, self.ysize, self.xsize)
        if self.nchannels > 1:
            s = s + (self.nchannels,)
        return s

    def open(self, file_name=None):
        if file_name is not None:
            self.file_name = file_name

        cmd = [
            'ffprobe',
            '-v', 'quiet',
            '-i', self.file_name,
            '-print_format', 'json',
            '-show_format',
            '-show_streams',
            '-select_streams', 'v:0'
        ]

        pipe = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE)
        self._probed_dict = json.loads(pipe.stdout.read().decode('utf-8'))
        pipe.communicate()

        if len(self._probed_dict) == 0 \
            or 'mp4' not in self._probed_dict['format']['format_name']:
                raise ValueError('Unrecognized format for FFMPEGWrapper')


    def zslice(self, start_frame, end_frame=None, dtype=None, copy=True):
        if end_frame is None:
            end_frame = start_frame + 1

        command = [
            'ffmpeg',
            '-i', self.file_name,
            '-f', 'image2pipe',
            '-vcodec', 'rawvideo',
            '-pix_fmt', 'gray' if 'gray' in self.pix_fmt else 'rgb24',
            '-'
        ]
        pipe = sp.Popen(command, stdout=sp.PIPE, stderr=sp.PIPE, bufsize=10**8)

        dt = np.dtype(self.dtype)

        i = 0
        while i < start_frame:
            pipe.stdout.read(self.ysize * self.xsize * dt.itemsize)
            i += 1

        shape = list(self.shape)
        shape[0] = end_frame - start_frame

        a = np.zeros(shape, dtype=self.dtype)

        shape[0] = 1

        while i < end_frame:
            raw = pipe.stdout.read(np.prod(shape) * dt.itemsize)
            a[i - start_frame] = \
                np.fromstring(raw, dtype=self.dtype).reshape(shape)
            i += 1

        pipe.communicate()

        if dtype is None:
            return a
        return a.astype(dtype)



import re
import math
import numpy as np

from functools import lru_cache

from .lcd_numbers import numbers, canvas_shape


def flatten(my_list):
    return [item for sublist in my_list for item in sublist]


def to_dtype(x, dtype):
    if x.dtype == dtype:
        return x
    if np.issubdtype(dtype, np.integer):
        np.rint(x, x)
    return x.astype(dtype, copy=False)


@lru_cache()
def squircle_alpha(height, width):
    squircle = np.zeros((height, width))
    ratio = width / height
    a = width // 2
    b = height // 2
    grid = np.vstack(np.meshgrid(np.linspace(0, b - 1, b),
                                 np.linspace(0, a - 1, a))).reshape(2, -1).T
    grid = grid.astype(np.int)
    N = max(a, b)
    ps = np.logspace(np.log10(2), np.log10(50), N)
    # ps = np.ones(N) * 2
    alpha = np.linspace(0, 1, N)

    if a > b:
        dra = a / N
        ras = np.arange(0, a, dra) + 1
        rbs = ras / ratio
        drb = dra / ratio
    else:
        drb = b / N
        rbs = np.arange(0, b, drb) + 1
        ras = rbs * ratio
        dra = drb * ratio

    counter = 0
    for y, x in grid:
        j = x / dra
        k = y / drb
        i = int(max(j, k))
        for n in range(0, N - i):
            ii = i + n
            try:
                p = ps[ii]
                ra = ras[ii]
                rb = rbs[ii]
            except IndexError:
                break

            constant = math.pow(x / ra, p) + math.pow(y / rb, p)

            if constant < 1:
                break

        squircle[y + b, x + a] = alpha[ii] ** 2
        counter += 1

    squircle[:b, a:] = np.flipud(squircle[b:, a:])
    squircle[:, :a] = np.fliplr(squircle[:, a:])

    squircle /= np.amax(squircle)
    squircle = 1 - squircle

    return squircle


def fuse_queue(q, dest, frame_shape, debug=False):
    """Fuse a queue of images along Y, optionally applying padding.

    Parameters
    ----------
    q : :py:class:`queue.Queue`
        A queue containing elements in the form ``[hyperslice, index, zfrom,
        offset_idx, top_left, overlaps]`` where `hyperslice` is a
        `numpy.ndarray`, `index` and `zfrom` are the tile index in the
        pandas dataframe and the starting frame in the original stack (these
        two are used for debugging purposes only), `offset_idx` is a
        tuple of :class:`slice` objects with the slice offset inside a single
        stack, `top_left` is a list specifying the image position in the form
        ``[Z, Y, X]``, `overlaps` is a :class:`pandas.DataFrame` specifying
        overlaps with adjacent tiles.
    frame_shape : tuple
        Shape of a stack plane (XY).
    dest : :class:`numpy.ndarray`
        Destination array.
    debug: bool
        Whether to overlay debug information (tile edges and numbers).
    """

    while True:
        slice, index_dbg, zfrom_dbg, sl, pos, overlaps = q.get()

        if slice is None:
            break

        z_from = pos[0]
        z_to = z_from + slice.shape[0]

        y_from = pos[1]
        y_to = y_from + slice.shape[-2]

        x_from = pos[2]
        x_to = x_from + slice.shape[-1]

        if overlaps is not None:
            z = np.array(flatten(overlaps[['Z_from', 'Z_to']].values))
            z = np.unique(z)
            z = np.sort(z)

            xy_weights = squircle_alpha(*frame_shape)

            z_list = list(zip(z, z[1::]))
            try:
                z_list += [(z[-1], None)]
            except IndexError:
                pass

            for zfrom, zto in z_list:
                sums = np.copy(xy_weights)
                condition = (overlaps['Z_from'] <= zfrom)
                if zto is not None:
                    condition = condition & (zto <= (overlaps['Z_to']))
                else:
                    condition = condition & (overlaps['Z_to'] >= z_to)

                for _, row in overlaps[condition].iterrows():
                    width = row.X_to - row.X_from
                    height = row.Y_to - row.Y_from
                    area = width * height
                    if not area:
                        continue

                    w = squircle_alpha(*frame_shape)[:height, :width]

                    if row.X_from == 0:
                        w = np.fliplr(w)
                    if row.Y_from == 0:
                        w = np.flipud(w)

                    xy_index = np.index_exp[row.Y_from:row.Y_to,
                                            row.X_from:row.X_to]
                    sums[xy_index] += w

                if zto is None:
                    slice_index = np.index_exp[zfrom:, ...]
                else:
                    slice_index = np.index_exp[zfrom:zto, ...]

                with np.errstate(invalid='ignore'):
                    factor = xy_weights / sums

                if sl is not None:
                    factor = factor[sl[-2::]]

                slice[slice_index] *= factor

        if debug:
            overlay_debug(slice, index_dbg, zfrom_dbg)
            slice[..., -2:, :] = 65000
            slice[..., -2:] = 65000

        output_roi_index = np.index_exp[z_from:z_to, ..., y_from:y_to,
                                        x_from:x_to]
        dest[output_roi_index] += slice

        q.task_done()


def overlay_debug(slice, index, z_from):
    cx = slice.shape[-1] // 2
    cy = slice.shape[-2] // 2 + 10
    x = cx - cx // 2
    for xstr in re.findall(r'\d+', index):
        for l in xstr:
            x_end = x + canvas_shape[1]
            try:
                slice[..., cy:cy + canvas_shape[0], x:x_end] = \
                    numbers[int(l)]
            except ValueError:
                break
            x = x_end + canvas_shape[1] // 2
        x = x_end + canvas_shape[1]

    cy += int(canvas_shape[0] * 1.4)
    for f in range(0, slice.shape[0]):
        x = cx
        xstr = str(z_from + f)
        for l in xstr:
            x_end = x + canvas_shape[1]
            ie = np.index_exp[f, ..., cy:cy + canvas_shape[0], x:x_end]
            if len(slice.shape) <= 3:
                ie = list(ie)[1::]
            try:
                slice[ie] = numbers[int(l)]
            except ValueError:
                break
            x = x_end + canvas_shape[1] // 2



major = 0
minor = 3
micro = 1

pre_release = ''
post_release = ''
dev_release = ''

__version__ = ''

if major != '':
    __version__ += '{}'.format(major)

if minor != '':
    __version__ += '.{}'.format(minor)

if micro != '':
    __version__ += '.{}'.format(micro)

if pre_release != '':
    __version__ += '{}'.format(pre_release)

if post_release != '':
    __version__ += '.post{}'.format(post_release)

if dev_release != '':
    __version__ += '.dev{}'.format(dev_release)



"""Parse input file names, compute tile coordinates."""

import os
import re
import logging

import json
import yaml

import numpy as np
import pandas as pd
import networkx as nx

from .inputfile import InputFile

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def parse_file_name(file_name):
    """Parse fields (stage coordinates) contained in `file_name`.

    Parameters
    ----------
    file_name : str
                The string to be parsed

    Returns
    -------
    x, y, z : int
        The parsed stage coordinates.
    """
    file_name = os.path.basename(file_name)
    m = re.search('^.*x_([-]?\d+).*y_([-]?\d+).*z_([-]?\d+).*', file_name)
    if m is None:
        m = re.search('^.*x_([-]?\d+).*y_([-]?\d+)', file_name)
    if m is None:
        m = re.search('^([-]?\d+)_([-]?\d+)_([-]?\d+)', file_name)
    if m is None:
        m = re.search('^([-]?\d+)_([-]?\d+)', file_name)
    if m is None:
        raise ValueError('Invalid name {}'.format(file_name))

    fields = []
    for i in range(1, 4):
        try:
            fields.append(int(m.group(i)))
        except IndexError:
            fields.append(0)

    logger.info('adding {} \tX={} Y={} Z={}'.format(file_name, *fields))
    return fields


class FileMatrix:
    """Data structures for a matrix of input files."""
    def __init__(self, input_path=None, ascending_tiles_x=True,
                 ascending_tiles_y=True):
        """
        Construct a FileMatrix object from a directory path or a .yml file
        produced by the stitcher. Tile ordering parameters need to be
        specified only if constructing from a directory, otherwise they are
        ignored.

        Parameters
        ----------
        input_path : str
                     input path (directory) or file (.yml)
        ascending_tiles_x : bool
            whether tiles are supposed to be read in ascending X order
        ascending_tiles_y : bool
            whether tiles are supposed to be read in ascending Y order
        """
        self.input_path = input_path

        self.data_frame = None
        """A :class:`pandas.DataFrame` object. Contains the following
        columns: `X`, `Y`, `Z`, `Z_end`, `xsize`, `ysize`, `nfrms`,
        `filename`."""

        self.ascending_tiles_x = ascending_tiles_x
        self.ascending_tiles_y = ascending_tiles_y

        self.name_array = None

        if input_path is None:
            return

        if os.path.isdir(input_path):
            self.load_dir(input_path)
        elif os.path.isfile(input_path):
            self.load_yaml(input_path)

    def load_dir(self, dir=None):
        """Look for files in `dir` recursively and populate data structures.

        Parameters
        ----------
        dir : path
        """
        if dir is None:
            dir = self.input_path

        if dir is None:
            return

        flist = []

        for root, dirs, files in os.walk(dir, followlinks=True):
            if os.path.basename(root):
                try:
                    self.parse_and_append(root, flist)
                    continue
                except (RuntimeError, ValueError):
                    pass

            for f in files:
                try:
                    self.parse_and_append(os.path.join(root, f), flist)
                    continue
                except (RuntimeError, ValueError):
                    pass

        data = {'X': flist[0::7], 'Y': flist[1::7], 'Z': flist[2::7],
                'nfrms': flist[3::7], 'ysize': flist[4::7],
                'xsize': flist[5::7], 'filename': flist[6::7]}
        df = pd.DataFrame(data)
        df = df.sort_values(['Z', 'Y', 'X'])

        if not self.ascending_tiles_x:
            df['X'] = (df['X'] - df['X'].max()).abs()

        if not self.ascending_tiles_y:
            df['Y'] = (df['Y'] - df['Y'].max()).abs()

        self.data_frame = df.set_index('filename')
        self.process_data_frame()

    def load_yaml(self, fname):
        logger.info('loading {}'.format(fname))
        with open(fname, 'r') as f:
            y = yaml.load(f)

        self.data_frame = pd.DataFrame(y['filematrix']).set_index('filename')
        self.data_frame = self.data_frame.sort_values(['Z', 'Y', 'X'])

        self.process_data_frame()
        self.input_path = fname

    def process_data_frame(self):
        df = self.data_frame

        xsize = df['X'].unique().size
        ysize = df['Y'].unique().size
        n_of_files = len(df.index)

        if xsize * ysize != n_of_files:
            msg = 'Mosaic is {}x{} tiles, but there are {} files!'.format(
                xsize, ysize, n_of_files)
            raise ValueError(msg)

        keys = ['X', 'Y', 'Z']
        df[keys] -= df[keys].min()

        cols = df.columns
        if 'Xs' in cols and 'Ys' in cols and 'Zs' in cols:
            for key in ['Xs', 'Ys', 'Zs']:
                df[key] -= df[key].min()

        df = df.sort_values(['Z', 'Y', 'X'])
        self.compute_end_pos()
        self.name_array = np.array(df.index.values).reshape(self.Ny, self.Nx)

    def parse_and_append(self, name, flist):
        try:
            fields = parse_file_name(name)
            with InputFile(name) as infile:
                fields.append(infile.nfrms)
                fields.append(infile.ysize)
                fields.append(infile.xsize)
            flist += fields
            flist.append(name)
        except (RuntimeError, ValueError):
            raise

    def save_to_yaml(self, filename, mode):
        keys = ['X', 'Y', 'Z', 'nfrms', 'xsize', 'ysize']
        abs_keys = ['Xs', 'Ys', 'Zs']
        for k in abs_keys:
            if k in self.data_frame.columns:
                keys.append(k)
        df = self.data_frame[keys].reset_index()
        j = json.loads(df.to_json(orient='records'))

        if mode == 'update':
            with open(filename, 'r') as f:
                y = yaml.load(f)

            y['filematrix'] = j

            logger.info('updating {}'.format(filename))
            with open(filename, 'w') as f:
                yaml.dump(y, f, default_flow_style=False)
        else:
            logger.info('writing {}'.format(filename))
            with open(filename, mode) as f:
                yaml.dump({'filematrix': j}, f, default_flow_style=False)

    def clear_absolute_positions(self):
        keys = ['Xs', 'Ys', 'Zs', 'Xs_end', 'Ys_end', 'Zs_end']
        for k in keys:
            try:
                del self.data_frame[k]
            except KeyError:
                pass

    def compute_end_pos(self):
        df = self.data_frame

        keys = ['X', 'Y', 'Z']
        sizes = ['xsize', 'ysize', 'nfrms']

        cols = df.columns
        if 'Xs' in cols and 'Ys' in cols and 'Zs' in cols:
            keys += ['Xs', 'Ys', 'Zs']
            sizes *= 2

        keys_end = [k + '_end' for k in keys]

        for ke, k, s in zip(keys_end, keys, sizes):
            df[ke] = df[k] + df[s]

    def compute_nominal_positions(self, px_size_z, px_size_xy):
        df = self.data_frame

        df['Xs'] = (df['X'] // px_size_xy).astype(np.int)
        df['Ys'] = (df['Y'] // px_size_xy).astype(np.int)

        df['Zs'] = (df['Z'] // px_size_z).astype(np.int)

        self.compute_end_pos()

    @property
    def Nx(self):
        return self.data_frame['X'].unique().size

    @property
    def Ny(self):
        return self.data_frame['Y'].unique().size

    @property
    def slices(self):
        """A slice is a group of tiles that share at least a `z` frame.

        Returns
        -------
        comp : generator
            A generator of graphs, one for each connected component of G,
            where G is the graph of tiles connected by at least a `z` frame.
        """
        G = nx.Graph()
        for index, row in self.data_frame.iterrows():
            G.add_node(index)

        for index, row in self.data_frame.iterrows():
            view = self.data_frame[
                (self.data_frame['Z'] <= row['Z'])
                & (self.data_frame['Z_end'] >= row['Z_end'])
                ]
            pairs = zip(view.index.values[::1], view.index.values[1::1])
            G.add_edges_from(pairs)
            G.add_edge((view.index.values[0]), view.index.values[-1])

        return nx.connected_component_subgraphs(G)

    @property
    def tiles_along_dir(self):
        """Groups of tiles to be stitched along a given direction.

        You need to send to this generator a tuple containing:
            - a list for sorting the :class:`pandas.DataFrame`, such as \
            :code:`['Z', 'Y', 'X']`

            - an axis for grouping, such as :code:`'Y'`

        Yields
        -------
        :class:`pandas.DataFrame`
            A group of tiles
        """
        for s in self.slices:
            got = yield
            view = self.data_frame.loc[s.nodes()].sort_values(
                got[0], ascending=True).groupby(got[1])
            for name, group in view:
                yield group

    @property
    def tiles_along_X(self):
        """Groups of tiles to be stitched along `X`.

        Equivalent to :attr:`~tiles_along_dir` having sent the following
        tuple: :code:`(['Z', 'X', 'Y'], 'Y')`

        Yields
        -------
        :class:`pandas.DataFrame`
            A group of tiles
        """
        g = self.tiles_along_dir
        next(g)
        yield g.send((['Z', 'X', 'Y'], 'Y'))
        yield from g

    @property
    def tiles_along_Y(self):
        """Groups of tiles to be stitched along `Y`.

        Equivalent to :attr:`~tiles_along_dir` having sent the following
        tuple: :code:`(['Z', 'Y', 'X'], 'X')`

        Yields
        -------
        :class:`pandas.DataFrame`
            A group of tiles
        """
        g = self.tiles_along_dir
        next(g)
        yield g.send((['Z', 'Y', 'X'], 'X'))
        yield from g

    @property
    def full_width(self):
        return self.data_frame['Xs_end'].max()

    @property
    def full_height(self):
        return self.data_frame['Ys_end'].max()

    @property
    def full_thickness(self):
        return self.data_frame['Zs_end'].max()



import numpy as np
import pandas as pd
import networkx as nx


def compute_shift_vectors(filematrix_dataframe, stitch_dataframe):
    sdf = stitch_dataframe
    fm_df = filematrix_dataframe

    sdf['px'] = 0
    sdf['py'] = 0
    sdf['pz'] = 0

    idx = sdf['axis'] == 1
    fm_df_idx = idx[idx].index
    sdf.loc[idx, 'px'] = sdf.loc[idx, 'dx']
    sdf.loc[idx, 'py'] = fm_df.loc[fm_df_idx, 'ysize'] - sdf.loc[idx, 'dy']
    sdf.loc[idx, 'pz'] = sdf.loc[idx, 'dz']

    idx = sdf['axis'] == 2
    fm_df_idx = idx[idx].index
    sdf.loc[idx, 'px'] = fm_df.loc[fm_df_idx, 'xsize'] - sdf.loc[idx, 'dy']
    sdf.loc[idx, 'py'] = -sdf.loc[idx, 'dx']
    sdf.loc[idx, 'pz'] = sdf.loc[idx, 'dz']


def compute_initial_guess(filematrix_dataframe, stitch_dataframe):
    fm_df = filematrix_dataframe
    sdf = stitch_dataframe.reset_index()

    G = nx.Graph()
    G.add_edges_from(((
        u, v, {'weight': w}) for u, v, w in
        np.c_[sdf['filename'], sdf['bname'], 1 - sdf['score']]))

    keys = ['Xs', 'Ys', 'Zs']
    for k in keys:
        fm_df[k] = 0

    sdf = sdf.reset_index().set_index('bname')

    cond = (fm_df['X'] == 0) & (fm_df['Y'] == 0) & (fm_df['Z'] == 0)
    top_left_corner = fm_df[cond].index[0]
    for edge in nx.bfs_edges(G, source=top_left_corner):
        btile = edge[1]
        parents = fm_df.loc[sdf.loc[[btile], 'filename']]
        temp_sdf = sdf.loc[[btile]].set_index('filename')
        temp = pd.DataFrame()
        temp['Xs'] = parents['Xs'] + temp_sdf['px']
        temp['Ys'] = parents['Ys'] + temp_sdf['py']
        temp['Zs'] = parents['Zs'] + temp_sdf['pz']
        temp['score'] = temp_sdf['score']
        fm_df.loc[btile, keys] = temp.apply(
            lambda x: np.average(x, weights=temp.loc[x.index, 'score']) if
            temp.loc[x.index, 'score'].sum() != 0 else np.average(x))

    fm_df[keys] -= fm_df[keys].min()
    fm_df[keys] = fm_df[keys].apply(np.round).astype(int)

    fm_df['Xs_end'] = fm_df['Xs'] + fm_df['xsize']
    fm_df['Ys_end'] = fm_df['Ys'] + fm_df['ysize']
    fm_df['Zs_end'] = fm_df['Zs'] + fm_df['nfrms']



import numpy as np

numbers = []
fill_value = 65535
canvas_shape = (50, 30)

canvas = np.zeros(canvas_shape, dtype=np.uint16)
cx = canvas.shape[1] // 2
cy = canvas.shape[0] // 2

# 0
a = np.zeros_like(canvas)
a[:2, :] = fill_value
a[-2:, :] = fill_value
a[:, -2:] = fill_value
a[:, :2] = fill_value
numbers.append(a)

# 1
a = np.zeros_like(canvas)
a[:, cx - 1:cx + 1] = fill_value
numbers.append(a)

# 2
a = np.zeros_like(canvas)
a[:2, :] = fill_value
a[:cy, -2:] = fill_value
a[cy - 1:cy + 1, :] = fill_value
a[cy:, :2] = fill_value
a[-2:, :] = fill_value
numbers.append(a)

# 3
a = np.zeros_like(canvas)
a[:2, :] = fill_value
a[cy - 1:cy + 1, :] = fill_value
a[-2:, :] = fill_value
a[:, -2:] = fill_value
numbers.append(a)

# 4
a = np.zeros_like(canvas)
a[:cy, :2] = fill_value
a[cy - 1:cy + 1, :] = fill_value
a[:, -2:] = fill_value
numbers.append(a)

# 5
a = np.zeros_like(canvas)
a[:2, :] = fill_value
a[:cy, :2] = fill_value
a[cy - 1:cy + 1, :] = fill_value
a[cy:, -2:] = fill_value
a[-2:, :] = fill_value
numbers.append(a)

# 6
a = np.zeros_like(canvas)
a[:2, :] = fill_value
a[cy - 1:cy + 1, :] = fill_value
a[-2:, :] = fill_value
a[cy:, -2:] = fill_value
a[:, :2] = fill_value
numbers.append(a)

# 7
a = np.zeros_like(canvas)
a[:2, :] = fill_value
a[:, -2:] = fill_value
numbers.append(a)

# 8
a = np.zeros_like(canvas)
a[:2, :] = fill_value
a[-2:, :] = fill_value
a[cy - 1:cy + 1, :] = fill_value
a[:, -2:] = fill_value
a[:, :2] = fill_value
numbers.append(a)

# 9
a = np.zeros_like(canvas)
a[:2, :] = fill_value
a[-2:, :] = fill_value
a[cy - 1:cy + 1, :] = fill_value
a[:, -2:] = fill_value
a[:cy, :2] = fill_value
numbers.append(a)



import pyfftw
import numpy as np


def normxcorr2_fftw(aslice, bframe):
    """Compute normalized cross correlation using fftw.

    .. DANGER::
        Input arrays might be overwritten by in-place transforms!

    Parameters
    ----------
    aslice : :class:`numpy.ndarray`
    bframe : :class:`numpy.ndarray`

    Returns
    -------
    :class:`numpy.ndarray`
    """
    ashape = aslice.shape
    b_old_shape = bframe.shape

    if not (aslice.dtype == np.float32):
        aslice = aslice.astype(np.float32)
    if not (bframe.dtype == np.float32):
        bframe = bframe.astype(np.float32)

    out_height = ashape[1] - b_old_shape[1] + 1
    out_width = ashape[2] - b_old_shape[2] + 1

    sums_b = np.sum(bframe)
    sums_b2 = np.sum(np.square(bframe))

    b1 = np.ones_like(bframe)
    bframe = np.pad(bframe, ((0, 0),
                             (0, ashape[1] - b_old_shape[1]),
                             (0, ashape[2] - b_old_shape[2])), 'constant')
    b1 = np.pad(b1, ((0, 0),
                     (0, ashape[1] - b_old_shape[1]),
                     (0, ashape[2] - b_old_shape[2] + 2)), 'constant')
    b1_real_input = b1[..., :ashape[2]]
    b1_complex_output = b1.view('complex64')

    # pad for in-place transform
    aslice = np.pad(aslice, ((0, 0), (0, 0), (0, 2)), 'constant')

    a_real_input = aslice[..., :ashape[2]]
    a_complex_output = aslice.view('complex64')

    fft_a2 = pyfftw.empty_aligned(a_complex_output.shape, dtype='complex64')
    fft_object = pyfftw.FFTW(np.square(a_real_input), fft_a2, axes=(1, 2),
                             flags=['FFTW_ESTIMATE'])
    fft_object.execute()

    # will overwrite aslice
    fft_object = pyfftw.FFTW(a_real_input, a_complex_output, axes=(1, 2),
                             flags=['FFTW_ESTIMATE'])
    fft_object.execute()

    # pad for in-place transform
    bframe = np.pad(bframe, ((0, 0), (0, 0), (0, 2)), 'constant')

    b_real_input = bframe[..., :ashape[2]]
    b_real_input[:] = bframe[..., :ashape[2]]
    b_complex_output = bframe.view('complex64')
    fft_object = pyfftw.FFTW(b_real_input, b_complex_output, axes=(1, 2),
                             flags=['FFTW_ESTIMATE'])
    fft_object.execute()

    fft_object = pyfftw.FFTW(
        b1_real_input, b1_complex_output, axes=(1, 2), flags=['FFTW_ESTIMATE'])
    fft_object.execute()

    # fftw performs unscaled transforms, therefore we need to rescale by the
    # frame area
    a_frame_area = np.array(ashape[1] * ashape[2], dtype=np.float32)

    conv = pyfftw.empty_aligned(ashape, dtype='float32')
    fft_object = pyfftw.FFTW(
        a_complex_output * np.conj(b_complex_output), conv, axes=(1, 2),
        flags=['FFTW_ESTIMATE'], direction='FFTW_BACKWARD')
    fft_object.execute()
    conv = conv[:, :out_height, :out_width] / a_frame_area

    fft_b1_conj = np.conj(b1_complex_output)
    sums_a = pyfftw.empty_aligned(ashape, dtype='float32')
    fft_object = pyfftw.FFTW(
        a_complex_output * fft_b1_conj, sums_a, axes=(1, 2),
        flags=['FFTW_ESTIMATE'], direction='FFTW_BACKWARD')
    fft_object.execute()
    sums_a = sums_a[:, :out_height, :out_width] / a_frame_area

    sums_a2 = pyfftw.empty_aligned(ashape, dtype='float32')
    fft_object = pyfftw.FFTW(
        fft_a2 * fft_b1_conj, sums_a2, axes=(1, 2),
        flags=['FFTW_ESTIMATE'], direction='FFTW_BACKWARD')
    fft_object.execute()
    sums_a2 = sums_a2[:, :out_height, :out_width] / a_frame_area

    A = np.array(b_old_shape[1] * b_old_shape[2], dtype=np.float32)

    num = conv - sums_b * sums_a / A
    denom = np.sqrt(np.abs(
        (sums_a2 - np.square(sums_a) / A) * (sums_b2 - np.square(sums_b) / A)))
    denom[denom == 0] = 1

    normxcorr = num / denom

    return normxcorr


def normxcorr2(aslice, bframe):
    ashape = aslice.shape
    bshape = bframe.shape

    out_height = ashape[1] - bshape[1] + 1
    out_width = ashape[2] - bshape[2] + 1

    b1 = np.ones_like(bframe)

    fft_a = np.fft.fft2(aslice)
    fft_a2 = np.fft.fft2(np.square(aslice))
    fft_b = np.fft.fft2(bframe, s=[ashape[1], ashape[2]])
    fft_b1 = np.fft.fft2(b1, s=[ashape[1], ashape[2]])

    conv = np.fft.ifft2(fft_a * np.conj(fft_b))
    conv = np.real(conv[:, :out_height, :out_width])
    sums_a = np.fft.ifft2(fft_a * np.conj(fft_b1))
    sums_a = np.real(sums_a[:, :out_height, :out_width])
    sums_a2 = np.fft.ifft2(fft_a2 * np.conj(fft_b1))
    sums_a2 = np.real(sums_a2[:, :out_height, :out_width])

    sums_b = np.sum(bframe)
    sums_b2 = np.sum(np.square(bframe))

    A = np.array(bshape[1] * bshape[2])

    num = conv - sums_b * sums_a / A
    denom = np.sqrt(
        (sums_a2 - np.square(sums_a) / A) * (sums_b2 - np.square(sums_b) / A))

    normxcorr = num / denom

    return normxcorr



import yaml
import numpy as np
import pandas as pd


class XcorrFileMatrix(object):
    def __init__(self):
        self.compute_average = False
        self.xcorr_options = None

        self.stitch_data_frame = None

    def load_yaml(self, fname):
        with open(fname, 'r') as f:
            y = yaml.load(f)

        attrs = ['ascending_tiles_x', 'ascending_tiles_y']

        for attr in attrs:
            setattr(self, attr, y['xcorr-options'][attr])

        self.stitch_data_frame = pd.DataFrame(y['xcorr'])
        self.stitch_data_frame = self.stitch_data_frame.rename(
            columns={'aname': 'filename'})
        self.stitch_data_frame = self.stitch_data_frame.set_index('filename')

        self.xcorr_options = y['xcorr-options']

    def aggregate_results(self, compute_average):
        sdf = self.stitch_data_frame.reset_index()

        if compute_average:
            view = sdf.groupby(['filename', 'bname', 'axis']).agg(
                lambda x: np.average(x, weights=sdf.loc[x.index, 'score']) if
                sdf.loc[x.index, 'score'].sum() != 0 else np.average(x))
        else:
            view = sdf.groupby(['filename', 'bname', 'axis']).agg(
                lambda x: sdf.loc[np.argmax(sdf.loc[x.index, 'score']), x.name]
            )

        view = view.reset_index()
        overlap_dict = {1: self.xcorr_options['overlap_v'],
                        2: self.xcorr_options['overlap_h']}

        view.dz -= self.xcorr_options['max_dz']
        for a in [1, 2]:
            indexes = (view['axis'] == a)
            view.loc[indexes, 'dy'] = (overlap_dict[a]
                                       + self.xcorr_options['max_dy']
                                       - view.loc[indexes, 'dy'])
        view.dx -= self.xcorr_options['max_dx']

        self.stitch_data_frame = view.set_index('filename')



import numpy as np
import pygmo as pg
import pandas as pd

from scipy.ndimage.interpolation import shift


class TileDisplacementLeastSquares:
    def __init__(self, xcorr_options, frame_shapes,
                 p_ab_1, score_1, p_ab_2, score_2):
        self.xcorr_options = xcorr_options
        self.frame_shapes = frame_shapes
        self.shape = p_ab_1.shape
        self.p_ab_1 = p_ab_1
        self.score_1 = score_1
        self.p_ab_2 = p_ab_2
        self.score_2 = score_2

    def get_bounds(self):
        n_of_tiles = np.prod(self.shape[0:2])
        xsize = self.shape[1]

        mins = [0, 0, 0]
        mins[0] = -self.xcorr_options['max_dz']
        mins[1] = None
        mins[2] = -self.xcorr_options['max_dx']
        mins = mins * n_of_tiles
        # overwrite None
        mins[1::3] = self.frame_shapes[:, 0] - (
            self.xcorr_options['overlap_v'] + self.xcorr_options['max_dy'])
        mins[:3] = [0, 0, 0]  # first tile is fixed
        # overwrite first row with horizontal shifts
        temp = [0, 0, 0]
        temp[0] = -self.xcorr_options['max_dz']
        temp[1] = -self.xcorr_options['max_dx']
        temp[2] = None
        mins[3:3 * xsize] = temp * (xsize - 1)
        # overwrite None
        mins[5:3 * xsize:3] = (self.frame_shapes[1:xsize, 1]
                               - self.xcorr_options['overlap_h']
                               - self.xcorr_options['max_dy'])

        maxs = [0, 0, 0]
        maxs[0] = self.xcorr_options['max_dz']
        maxs[1] = None
        maxs[2] = self.xcorr_options['max_dx']
        maxs = maxs * n_of_tiles
        # overwrite None
        maxs[1::3] = (self.frame_shapes[:, 0] - self.xcorr_options['overlap_v']
                      + self.xcorr_options['max_dy'])
        maxs[:3] = [0, 0, 0]  # first tile is fixed
        # overwrite first row with horizontal shifts
        temp[0] = self.xcorr_options['max_dz']
        temp[1] = self.xcorr_options['max_dx']
        temp[2] = None
        maxs[3:3 * xsize] = temp * (xsize - 1)
        # overwrite None
        maxs[5:3 * xsize:3] = (self.frame_shapes[1:xsize, 1]
                               - self.xcorr_options['overlap_h']
                               + self.xcorr_options['max_dy'])

        mins = [int(x) for x in mins]
        maxs = [int(x) for x in maxs]

        bounds = (mins, maxs)
        return bounds

    def fitness(self, x):
        t = decision_vector_to_tile_coords(x, self.shape)

        sums = 0

        # add shifts along x
        shifted = shift(t, (0, -1, 0), cval=np.NaN)
        norms = np.linalg.norm(
            np.nan_to_num((shifted - t - self.p_ab_2)), axis=-1)
        sums += np.sum(self.score_2 * norms**2)

        # add shifts along y
        shifted = shift(t, (-1, 0, 0), cval=np.NaN)
        norms = np.linalg.norm(
            np.nan_to_num((shifted - t - self.p_ab_1)), axis=-1)
        sums += np.sum(self.score_1 * norms**2)

        return [sums]

    def gradient(self, x):
        return pg.estimate_gradient_h(lambda x: self.fitness(x), x)


def decision_vector_to_tile_coords(x, shape):
    """Covert decision vector `x` to tile coordinates.

    Parameters
    ----------
    x : :class:`numpy.ndarray`
        Decision vector. First row: displacement vector from adjacent left
        tile. Other rows: displacement vector from adjacent top tile.
    shape : list
        `[ysize, xsize, 3]` where `ysize` and `xsize` are the mosaic size.

    Returns
    -------

    """
    t = x.reshape(shape)
    t[0, ...] = np.cumsum(t[0, ...], axis=0)
    t = np.cumsum(t, axis=0)
    return t


def absolute_position_global_optimization(df, sdf, xcorr_options):
    """Perform global optimization to adjust tile absolute positions.

    The passed dataframe must come with an initial guess of absolute positions.

    Parameters
    ----------
    df : :attr:`.FileMatrix.data_frame`
    sdf : :attr:`.FileMatrix.stitch_data_frame`
    """
    # mosaic size
    xsize = df['X'].unique().size
    ysize = df['Y'].unique().size

    idx = df.sort_values(['Z', 'Y', 'X']).index
    frame_shapes = np.array(df.loc[idx, ['ysize', 'xsize']])

    # shifts along Y
    temp = sdf.loc[idx, ['pz', 'py', 'px', 'axis', 'score']]
    try:
        temp = temp[temp['axis'] == 1].loc[idx]
        p_ab_1 = np.array(temp[['pz', 'py', 'px']]).reshape(ysize, xsize, 3)
        score_1 = np.nan_to_num(np.array(temp['score']).reshape(ysize, xsize))
    except KeyError:
        p_ab_1 = np.zeros((ysize, xsize, 3))
        score_1 = np.zeros((ysize, xsize))

    shape = p_ab_1.shape

    # shifts along X
    temp = sdf.loc[idx, ['pz', 'py', 'px', 'axis', 'score']]
    try:
        temp = temp[temp['axis'] == 2].loc[idx]
        p_ab_2 = np.array(temp[['pz', 'py', 'px']]).reshape(ysize, xsize, 3)
        score_2 = np.nan_to_num(np.array(temp['score']).reshape(ysize, xsize))
    except KeyError:
        p_ab_2 = np.zeros((ysize, xsize, 3))
        score_2 = np.zeros((ysize, xsize))

    # initial decision vector
    # first row: displacement vector from adjacent left tile
    # other rows: displacement vector from adjacent top tile
    x0 = np.nan_to_num(np.roll(p_ab_1, 1, axis=0))
    x0[0, ...] = np.nan_to_num(np.roll(p_ab_2[0, ...], 1, axis=0))

    prob = TileDisplacementLeastSquares(xcorr_options, frame_shapes,
                                        p_ab_1, score_1, p_ab_2, score_2)

    archi = pg.archipelago(n=0)

    pop = pg.population(prob=prob)
    pop.push_back(x0.flatten())
    for i in range(0, 8):
        # each algo gets a different seed
        algo = pg.algorithm(
            pg.simulated_annealing(Ts=10., Tf=1e-5, n_T_adj=10))
        if not i:
            algo.set_verbosity(500)
        archi.push_back(algo=algo, pop=pop)

    archi.evolve(2)
    archi.wait_check()

    fs = archi.get_champions_f()
    xs = archi.get_champions_x()
    champion = xs[0]
    current_min = fs[0][0]
    for x, f in zip(xs, fs):
        if f < current_min:
            current_min = f[0]
            champion = x

    print(current_min)

    # rebuild tile coordinates starting from decision vector x
    t = decision_vector_to_tile_coords(champion, shape).reshape(-1, 3)

    keys = ['Zs', 'Ys', 'Xs']
    ret_df = pd.DataFrame(data=t, index=idx, columns=keys)

    df[keys] = ret_df
    df[keys] -= df[keys].min()
    df[keys] = df[keys].apply(np.round).astype(int)

    df['Xs_end'] = df['Xs'] + df['xsize']
    df['Ys_end'] = df['Ys'] + df['ysize']
    df['Zs_end'] = df['Zs'] + df['nfrms']



"""API to query an arbitrary region in the stitched volume."""

import os.path
import logging
import threading

from queue import Queue
from functools import lru_cache

import numpy as np

from .overlaps import Overlaps
from ..inputfile import InputFile
from ..filematrix import FileMatrix
from .fuse import fuse_queue, to_dtype

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class VirtualFusedVolume:
    """An API to query arbitrary regions in the stitched volume.

    Example usage:

    >>> vfv = VirtualFusedVolume('stitch.yml')
    >>> vfv.shape
    (208, 3, 4533, 2487)

    Axis order is ZCYX.

    >>> subvolume = vfv[40, ..., 1000:1500, 2000:2400]

    """
    def __init__(self, file_or_matrix):
        if isinstance(file_or_matrix, str):
            self.path, _ = os.path.split(file_or_matrix)
            self.fm = FileMatrix(file_or_matrix)
        else:
            self.fm = file_or_matrix
            self.path, _ = os.path.split(self.fm.input_path)

        absolute_positions_found = True
        abs_keys = ['Xs', 'Ys', 'Zs']
        for k in abs_keys:
            if not k in self.fm.data_frame.columns:
                absolute_positions_found = False
                break

        if absolute_positions_found:
            self.ov = Overlaps(self.fm)
        else:
            self.fm.compute_nominal_positions(1, 1)
            self.ov = None

        self._debug = False

        infile = os.path.join(self.path, self.fm.data_frame.iloc[0].name)
        with InputFile(infile) as f:
            self.temp_shape = list(f.shape)
            self.dtype = f.dtype
            self.nchannels = f.nchannels

    @property
    def overlay_debug_enabled(self):
        """Whether to overlay debug information (tile edges and numbers).

        Defaults to `False`."""
        return self._debug

    @overlay_debug_enabled.setter
    def overlay_debug_enabled(self, value):
        self._debug = value

    @property
    @lru_cache()
    def shape(self):
        """Shape of the whole stitched volume.

        Axis order is ZCYX.
        """
        thickness = self.fm.full_thickness

        infile = os.path.join(self.path, self.fm.data_frame.iloc[0].name)
        with InputFile(infile) as f:
            output_shape = list(f.shape)

        output_shape[0] = thickness
        output_shape[-2] = self.fm.full_height
        output_shape[-1] = self.fm.full_width

        return tuple(output_shape)

    def __getitem__(self, item):

        item = np.index_exp[item]  # ensure item is a tuple

        # ensure all items are slice objects
        myitem = []
        flip_axis = []
        for i in item:
            if isinstance(i, int):
                start = i
                stop = i + 1
                step = 1
            elif i is Ellipsis:
                for _ in range(0, len(self.shape) - len(item) + 1):
                    myitem.append(slice(0, self.shape[len(myitem)], 1))
                    flip_axis.append(1)
                continue
            elif isinstance(i, slice):
                start = i.start
                stop = i.stop
                step = i.step if i.step is not None else 1
            else:
                raise TypeError("Invalid type: {}".format(type(i)))

            curr_max = self.shape[len(myitem)]
            if start is None:
                start = 0 if step > 0 else curr_max
            elif start < 0:
                start += curr_max
            elif start > curr_max:
                start = curr_max

            if stop is None:
                stop = curr_max if step > 0 else 0
            elif stop < 0:
                stop += curr_max
            elif stop > curr_max:
                stop = curr_max

            if step < 0:
                start += abs(step)
                stop += abs(step)

            flip_axis.append(1 if step > 0 else -1)

            if start < stop:
                myitem.append(slice(start, stop, abs(step)))
            else:
                myitem.append(slice(stop, start, abs(step)))

        for _ in range(0, len(self.shape) - len(myitem)):
            myitem.append(slice(0, self.shape[len(myitem)], 1))

        if len(myitem) != len(self.shape):
            raise IndexError('Too many indices for array')

        zmin = myitem[0].start
        zmax = myitem[0].stop

        ymin = myitem[-2].start
        ymax = myitem[-2].stop

        xmin = myitem[-1].start
        xmax = myitem[-1].stop

        output_shape = [
            (it.stop - it.start) // it.step for it in myitem
        ]
        if 0 in output_shape:
            return np.array([], dtype=self.dtype)

        dtype = np.float32 if self.ov is not None else self.dtype
        fused = np.zeros(output_shape, dtype=dtype)

        q = Queue(maxsize=20)

        t = threading.Thread(
            target=fuse_queue,
            args=(q, fused, self.temp_shape[-2::], self._debug))
        t.start()

        sl = myitem[:]

        X_min = np.array([myitem[i].start for i in [0, -2, -1]])
        X_stop = np.array([myitem[i].stop for i in [0, -2, -1]])
        steps = np.array([myitem[i].step for i in [0, -2, -1]])

        df = self.fm.data_frame
        df = df[
            (df['Zs_end'] > zmin) & (df['Zs'] < zmax)
            & (df['Ys_end'] > ymin) & (df['Ys'] < ymax)
            & (df['Xs_end'] > xmin) & (df['Xs'] < xmax)
        ]

        if df.shape[0] == 1:
            index = df.iloc[0].name
            with InputFile(os.path.join(self.path, index)) as f:
                try:
                    f.wrapper.deep_copy_enabled = True
                except AttributeError:
                    pass
                logger.info('opening {}\t{}'.format(index, sl))
                sl_a = f[item]
            return sl_a

        for index, row in df.iterrows():
            Xs = np.array([row.Zs, row.Ys, row.Xs])
            xsize = np.array([row.nfrms, row.ysize, row.xsize])

            xto = X_stop - Xs
            xto[xto > xsize] = xsize[xto > xsize]
            xfrom = X_min - Xs
            xfrom[xfrom < 0] = 0
            xfrom = xfrom + (Xs + xfrom - X_min) % steps

            for i in [0, -2, -1]:
                sl[i] = slice(xfrom[i], xto[i], steps[i])

            z_from = sl[0].start
            z_to = sl[0].stop

            x_from = np.array([sl[i].start for i in [0, -2, -1]])

            logger.info('opening {}\t{}'.format(index, sl))
            with InputFile(os.path.join(self.path, index)) as f:
                sl_a = np.copy(f[tuple(sl)]).astype(dtype)
            sl_a.shape = ([1 for _ in range(0, len(sl) - len(sl_a.shape))]
                          + list(sl_a.shape))

            Top_left = Xs + x_from
            top_left = (Top_left - X_min) // steps

            if self.ov is None:
                overlaps = None
            else:
                overlaps = self.ov[index]
                overlaps = overlaps.loc[
                    (overlaps['Z_from'] <= z_to) & (overlaps['Z_to'] >= z_from)
                    ].copy()

                overlaps['Z_from'] -= z_from
                overlaps['Z_to'] -= z_from
                overlaps['Z_to'] /= abs(sl[0].step)
                overlaps['Z_to'] = overlaps['Z_to'].apply(np.round).astype(int)

                overlaps.loc[overlaps['Z_from'] < 0, 'Z_from'] = 0

            q.put([sl_a, index, z_from, tuple(sl), top_left, overlaps])

        q.put([None, None, None, None, None, None])  # close queue

        t.join()  # wait for fuse thread to finish

        fused = to_dtype(fused, self.dtype)

        ie = [slice(None, None, flip) for flip in flip_axis]
        fused = fused[ie]

        return np.squeeze(fused)



from setuptools import setup, find_packages


version = {}
with open("zetastitcher/version.py") as fp:
    exec(fp.read(), version)
__version__ = version['__version__']

setup(
    name='zetastitcher',
    version=__version__,
    description='Stitch large volumetric images (3D stacks)',
    long_description='ZetaStitcher is a tool designed to stitch large '
                     'volumetric images such as those produced by Light-Sheet '
                     'Fluorescence Microscopes.',
    author='Giacomo Mazzamuto',
    author_email='mazzamuto@lens.unifi.it',
    license='GPLv3+',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    # What does your project relate to?
    keywords='stitch, microscopy',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'coloredlogs',
        'networkx',
        'numpy',
        'pandas',
        'psutil',
        'pyfftw',
        'pygmo>=2',
        'pyyaml',
        'scipy',
        'scikit-image',
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': [
            'pip-tools',
        ],
        'dcimg': [
            'dcimg>=0.4.0'
        ],
        'doc': [
            'numpydoc',
            'sphinx',
            'sphinx_rtd_theme',
        ],
        'test': ['mock'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
    },

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'stitch-align = zetastitcher.runner:main',
            'stitch-fuse = zetastitcher.fuser.__main__:main',
        ],

    },
)
